package dev.fyloz.colorrecipesexplorer.config.initializers

import dev.fyloz.colorrecipesexplorer.config.annotations.RequireDatabase
import dev.fyloz.colorrecipesexplorer.model.Mix
import dev.fyloz.colorrecipesexplorer.model.MixMaterial
import dev.fyloz.colorrecipesexplorer.service.MixService
import dev.fyloz.colorrecipesexplorer.utils.merge
import mu.KotlinLogging
import org.springframework.context.annotation.Configuration
import java.util.*

@Configuration
@RequireDatabase
class MixInitializer(
    private val mixService: MixService
) : AbstractInitializer() {
    private val logger = KotlinLogging.logger {}

    override fun initialize() {
        logger.debug("Executing mix initializer...")
        fixAllPositions()
    }

    private fun fixAllPositions() {
        logger.debug("Validating mix materials positions...")

        mixService.getAll()
            .filter { mix -> mix.mixMaterials.any { it.position == 0 } }
            .forEach(this::fixMixPositions)

        logger.debug("Mix materials positions are valid!")
    }

    private fun fixMixPositions(mix: Mix) {
        val maxPosition = mix.mixMaterials.maxOf { it.position }

        logger.warn("Mix ${mix.id} (${mix.mixType.name}, ${mix.recipe.name}) has invalid positions:")

        val invalidMixMaterials: Collection<MixMaterial> = with(mix.mixMaterials.filter { it.position == 0 }) {
            if (maxPosition == 0 && this.size > 1) {
                orderMixMaterials(this)
            } else {
                this
            }
        }

        val fixedMixMaterials = increaseMixMaterialsPosition(invalidMixMaterials, maxPosition + 1)
        val updatedMixMaterials = mix.mixMaterials.merge(fixedMixMaterials)

        with(mix.copy(mixMaterials = updatedMixMaterials.toMutableSet())) {
            mixService.update(this)
        }
    }

    private fun increaseMixMaterialsPosition(mixMaterials: Iterable<MixMaterial>, firstPosition: Int) =
        mixMaterials
            .mapIndexed { index, mixMaterial -> mixMaterial.copy(position = firstPosition + index) }
            .onEach {
                logger.info("\tPosition of material ${it.material.id} (${it.material.name}) has been set to ${it.position}")
            }

    private fun orderMixMaterials(mixMaterials: Collection<MixMaterial>) =
        LinkedList(mixMaterials).apply {
            while (this.peek().material.materialType?.usePercentages == true) {
                // The first mix material can't use percents, so move it to the end of the queue
                val pop = this.pop()
                this.add(pop)
                logger.debug("\tMaterial ${pop.material.id} (${pop.material.name}) uses percents, moving to the end of the queue")
            }
        }
}